unit DateEd97;

(******************************************************************************
TDateEdit97

A date edit field with drop down calendar.

PROPERTIES:

Date - TDateTime that contains the date value of the control.

ValidDateColor - The color that "valid dates" will be rendered.

METHODS:

procedure AddValidDate - Adds a datetime value to a list of "valid dates"
maintained by the control. These dates will be drawn in the color specified by
ValidDateColor.

procedure ClearValidDates - Clears all "valid dates" from the list.

function DateInList - Checks if the specified date is in the list of "valid dates".

EVENTS:

OnDateChange - Triggered whenever the Date property is updated.
*********************************************)

interface

uses
  SysUtils, WinTypes, WinProcs, Messages, Classes, Graphics, Controls,
  Forms, StdCtrls, Buttons, IniFiles, Db, DbTables, CalPop;

type
  PTDateTime = ^TDateTime;
  TDay = string[2];
  TDayArray = array[1..7] of TDay;
  TDayOfWeek = 1..7;
  TDayName = class(TPersistent)
  private
    FMonday: TDay;
    FTuesday: TDay;
    FWednesday: TDay;
    FThursday: TDay;
    FFriday: TDay;
    FSaturday: TDay;
    FSunday: TDay;
  protected
  public
    constructor Create;
    destructor Destroy; override;
  published
    property Monday: TDay read FMonday write FMonday;
    property Tuesday: TDay read FTuesday write FTuesday;
    property Wednesday: TDay read FWednesday write FWednesday;
    property Thursday: TDay read FThursday write FThursday;
    property Friday: TDay read FFriday write FFriday;
    property Saturday: TDay read FSaturday write FSaturday;
    property Sunday: TDay read FSunday write FSunday;
  end;

  TMonth = string[15];
  TMonthArray = array[1..12] of TMonth;
  TMonthName = class(TPersistent)
  private
    FJanuary: TMonth;
    FFebruary: TMonth;
    FMarch: TMonth;
    FApril: TMonth;
    FMay: TMonth;
    FJune: TMonth;
    FJuly: TMonth;
    FAugust: TMonth;
    FSeptember: TMonth;
    FOctober: TMonth;
    FNovember: TMonth;
    FDecember: TMonth;
  protected
  public
    constructor Create;
    destructor Destroy; override;
  published
    property January: TMonth read FJanuary write FJanuary;
    property February: TMonth read FFebruary write FFebruary;
    property March: TMonth read FMarch write FMarch;
    property April: TMonth read FApril write FApril;
    property May: TMonth read FMay write FMay;
    property June: TMonth read FJune write FJune;
    property July: TMonth read FJuly write FJuly;
    property August: TMonth read FAugust write FAugust;
    property September: TMonth read FSeptember write FSeptember;
    property October: TMonth read FOctober write FOctober;
    property November: TMonth read FNovember write FNovember;
    property December: TMonth read FDecember write FDecember;
  end;

  TDateButton = class(TBitBtn)
  private
  protected
     procedure Click; override;
  public
  published
  end;

  TDateEdit97 = class(TEdit)
  private
    FCanClear : boolean;
{    hBitmap: HBitmap;}
    FBtnClear : tBitBtn;
    FBtnPopup: TDateButton;
    FDate: TDateTime;
    FOnDateChange: TNotifyEvent;
    FDayName: TDayName;
    FMonthName: TMonthName;
    FStartOfWeek: TDayOfWeek;
    FValColor: TColor;
    lstDates: TList;
    Token: Integer;
    procedure SetToken;
    procedure SelectToken;
    procedure SetSeperators;
    procedure SetStartOfWeek(Value: TDayOfWeek);
  protected
    nSep1, nSep2: Integer;
    procedure SetCanClear (value : boolean);
    procedure WMSize(var message: TWMSize); message WM_SIZE;
    function GetDate: TDateTime;
    procedure SetDate(dtArg: TDateTime);
    function GetDayArray: TDayArray;
    function GetMonthArray: TMonthArray;
    procedure KeyPress(var Key: char); override;
    procedure MouseUp(Button: TMouseButton; ShiftState: TShiftState; X, Y: Integer); override;
    procedure DoExit; override;
    procedure DoEnter; override;
    procedure ClearDate (Sender: TObject);
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure CreateParams(var Params: TCreateParams); override;
    property Date: TDateTime read GetDate write SetDate;
    property DayArray: TDayArray read GetDayArray;
    property MonthArray: TMonthArray read GetMonthArray;
    function DateInList(dt: TDateTime): Boolean;
    procedure AddValidDate(dt: TDateTime);
    procedure ClearValidDates;
  published
    property CanClear: boolean read FCanClear write SetCanClear default true;
    property OnDateChange: TNotifyEvent read FOnDateChange write FOnDateChange;
    property DayNames: TDayName read FDayName write FDayName;
    property MonthNames: TMonthName read FMonthName write FMonthName;
    property StartOfWeek: TDayOfWeek read FStartOfWeek write SetStartOfWeek default 1;
    property ValidDateColor: TColor read FValColor write FValColor default clMaroon;
  end;

  TDbDateEdit97 = class(TDateEdit97)
  private
    FDataLink: TFieldDataLink;
    FCanvas: TControlCanvas;
    FAlignment: TAlignment;
    FFocused: Boolean;
    FTextMargin: Integer;
    procedure DataChange(Sender: TObject);
    procedure EditingChange(Sender: TObject);
    function GetDataField: string;
    function GetDataSource: TDataSource;
    function GetField: TField;
    function GetReadOnly: Boolean;
    procedure SetDataField(const Value: string);
    procedure SetDataSource(Value: TDataSource);
    procedure SetFocused(Value: Boolean);
    procedure SetReadOnly(Value: Boolean);
    procedure UpdateData(Sender: TObject);
    procedure WMCut(var message: TMessage); message WM_CUT;
    procedure WMPaste(var message: TMessage); message WM_PASTE;
    procedure CMEnter(var message: TCMEnter); message CM_ENTER;
    procedure CMExit(var message: TCMExit); message CM_EXIT;
  protected
    procedure SetDate(dtArg: TDateTime);
    procedure Change; override;
    procedure KeyPress(var Key: char); override;
    procedure KeyDown(var Key: Word; Shift: TShiftState); override;
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
    procedure Click; override;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    property Field: TField read GetField;
  published
    property DataField: string read GetDataField write SetDataField;
    property DataSource: TDataSource read GetDataSource write SetDataSource;
    property ReadOnly: Boolean read GetReadOnly write SetReadOnly default False;
  end;

procedure Register;

var
  frmCalendar: TfrmCalPop;

implementation

{$R *.RES}

{-------------------------------------------------------------}
{---------------------- TDayName -----------------------------}
{-------------------------------------------------------------}
constructor TDayName.Create;
begin
  inherited Create;
  FSunday := ShortDayNames[1];
  FMonday := ShortDayNames[2];
  FTuesday := ShortDayNames[3];
  FWednesday := ShortDayNames[4];
  FThursday := ShortDayNames[5];
  FFriday := ShortDayNames[6];
  FSaturday := ShortDayNames[7];
end;

destructor TDayName.Destroy;
begin
  inherited Destroy;
end;

{-------------------------------------------------------------}
{---------------------- TMonthName ---------------------------}
{-------------------------------------------------------------}
constructor TMonthName.Create;
var
  ADate: TDateTime;
begin
  inherited Create;
  FJanuary := FormatDateTime( 'mmmm', EncodeDate(1998, 01, 01));
  FFebruary := FormatDateTime( 'mmmm', EncodeDate(1998, 02, 01));
  FMarch := FormatDateTime( 'mmmm', EncodeDate(1998, 03, 01));
  FApril := FormatDateTime( 'mmmm', EncodeDate(1998, 04, 01));
  FMay := FormatDateTime( 'mmmm', EncodeDate(1998, 05, 01));
  FJune := FormatDateTime( 'mmmm', EncodeDate(1998, 06, 01));
  FJuly := FormatDateTime( 'mmmm', EncodeDate(1998, 07, 01));
  FAugust := FormatDateTime( 'mmmm', EncodeDate(1998, 08, 01));
  FSeptember := FormatDateTime( 'mmmm', EncodeDate(1998, 09, 01));
  FOctober := FormatDateTime( 'mmmm', EncodeDate(1998, 10, 01));
  FNovember := FormatDateTime( 'mmmm', EncodeDate(1998, 11, 01));
  FDecember := FormatDateTime( 'mmmm', EncodeDate(1998, 12, 01));
end;

destructor TMonthName.Destroy;
begin
  inherited Destroy;
end;

{-------------------------------------------------------------}
{---------------------- TDateButton --------------------------}
{-------------------------------------------------------------}
procedure TDateButton.Click;
var
  editParent: TDateEdit97;
begin
  editParent := TDateEdit97(Parent);
  editParent.Click;
  frmCalendar := TfrmCalPop.Create(editParent);
  frmCalendar.ShowModal;
  frmCalendar.Free;
  inherited Click;
end;

{-------------------------------------------------------------}
{---------------------- TDateEdit97 --------------------------}
{-------------------------------------------------------------}
constructor TDateEdit97.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  Token := 1;
  FCanClear := true;
  FDate := 0;
  FStartOfWeek := 1;
  FDayName := TDayName.Create;
  FMonthName := TMonthName.Create;
  FValColor := clBlue;
  FBtnClear := tBitBtn.Create (Self);
  FBtnClear.Visible := True;
  FBtnClear.Parent := Self;
  FBtnClear.Glyph.Handle := LoadBitmap(hInstance, 'BTNCLEAR');
  FBtnClear.OnClick := ClearDate;
  FBtnPopup := TDateButton.Create(Self);
  FBtnPopup.Visible := True;
  FBtnPopup.Parent := Self;
  FBtnPopup.Glyph.Handle := LoadBitmap(hInstance, 'BTNPOPUP');
  ControlStyle := ControlStyle - [csSetCaption];
  lstDates := TList.Create;
end;

procedure TDateEdit97.CreateParams(var Params: TCreateParams);
begin
  inherited CreateParams(Params);
  Params.Style := Params.Style or WS_CLIPCHILDREN;
end;

destructor TDateEdit97.Destroy;
begin
  FBtnClear.Free;
  FBtnPopup.Free;

  FDayName.Free;
  FMonthName.Free;
  ClearValidDates;
  lstDates.Free;
  inherited Destroy;
end;

procedure TDateEdit97.WMSize(var message: TWMSize);
begin
  FBtnPopup.Height := Height;
  FBtnPopup.Width := Height;
  FBtnPopup.Left := Width - Height;
  FBtnPopup.Refresh;

  FBtnClear.Height := Height;
  FBtnClear.Width := Height;
  FBtnClear.Left := Width - Height*2;
  FBtnClear.Refresh;
end;

function TDateEdit97.GetDate: TDateTime;
begin
  GetDate := FDate;
end;

procedure TDateEdit97.SetDate(dtArg: TDateTime);
var
  FormattedDate: string;
begin
  if FDate <> dtArg then
  begin
    FDate := dtArg;
    Modified := True;
    if FDate = 0 then
      Text := ''
    else
      Text := FormatDateTime(ShortDateFormat, FDate);

    if Assigned(FOnDateChange) then FOnDateChange(Self);
  end;
end;

function TDateEdit97.GetDayArray: TDayArray;
var
  FDayArray: TDayArray;
begin
  FDayArray[1] := DayNames.FMonday;
  FDayArray[2] := DayNames.FTuesday;
  FDayArray[3] := DayNames.FWednesday;
  FDayArray[4] := DayNames.FThursday;
  FDayArray[5] := DayNames.FFriday;
  FDayArray[6] := DayNames.FSaturday;
  FDayArray[7] := DayNames.FSunday;
  Result := FDayArray;
end;

function TDateEdit97.GetMonthArray: TMonthArray;
var
  FMonthArray: TMonthArray;
begin
  FMonthArray[01] := MonthNames.FJanuary;
  FMonthArray[02] := MonthNames.FFebruary;
  FMonthArray[03] := MonthNames.FMarch;
  FMonthArray[04] := MonthNames.FApril;
  FMonthArray[05] := MonthNames.FMay;
  FMonthArray[06] := MonthNames.FJune;
  FMonthArray[07] := MonthNames.FJuly;
  FMonthArray[08] := MonthNames.FAugust;
  FMonthArray[09] := MonthNames.FSeptember;
  FMonthArray[10] := MonthNames.FOctober;
  FMonthArray[11] := MonthNames.FNovember;
  FMonthArray[12] := MonthNames.FDecember;
  Result := FMonthArray;
end;

procedure TDateEdit97.DoEnter;
begin
  inherited DoEnter;
  Token := 1;
  SetSeperators;
  SelectToken;
end;

procedure TDateEdit97.DoExit;
begin
  inherited DoExit;
  try
    FDate := StrToDate(Text);
  except
    If Length(Text)<>0
      then SetFocus
      else FDate := 0;
  end;
end;

(*********************************************
Is the supplied data in the date list?
*********************************************)
function TDateEdit97.DateInList(dt: TDateTime): Boolean;
var
  pDate: PTDateTime;
  I: Integer;
begin
  Result := False;
  for I := 0 to lstDates.Count - 1 do
  begin
    pDate := lstDates[I];
    if pDate^ = dt then
    begin
      Result := True;
      Break;
    end;
  end;
end;

(*********************************************
Maintain list of valid dates.
*********************************************)
procedure TDateEdit97.AddValidDate(dt: TDateTime);
var
  pDate: PTDateTime;
begin
  New(pDate);
  pDate^ := dt;
  lstDates.Add(PDate);
end;

procedure TDateEdit97.ClearValidDates;
var
  pDate: PTDateTime;
begin
  while lstDates.Count > 0 do
  begin
    pDate := lstDates[0];
    Dispose(pDate);
    lstDates.Delete(0);
  end;
end;

procedure TDateEdit97.KeyPress(var Key: Char);
begin
  if ((Key < '0') or (Key > '9')) and (Key <> DateSeparator) and (Key <> #8) then
    Key := #0
  else
  if Key = DateSeparator then
  begin
    if Token < 3 then
    begin
      Inc(Token);
      SetSeperators;
      SelectToken;
      Key := #0;
    end
    else
      Key := #0;
  end
  else
     inherited KeyPress(Key);
end;

(*********************************************
Determine which token the user is on and highlight
the entire text of that token.
*********************************************)
procedure TDateEdit97.MouseUp(Button: TMouseButton; ShiftState: TShiftState; X, Y: Integer);
begin
  SetToken;
  SelectToken;
  inherited MouseUp(Button, ShiftState, X, Y);
end;

procedure TDateEdit97.SetStartOfWeek(Value: TDayOfWeek);
begin
  if FStartOfWeek <> Value then FStartOfWeek := Value;
end;

(*********************************************
Set the positions of the seperators in text.
*********************************************)
procedure TDateEdit97.SetSeperators;
var
  I: Integer;
begin
  nSep1 := Pos(DateSeparator, Text);
  for I := nSep1 + 1 to Length(Text) do begin
    if Text[I] = DateSeparator then
    begin
      nSep2 := i;
      Break;
    end;
  end;
end;

(*********************************************
Determine which token the cursor is over;
*********************************************)
procedure TDateEdit97.SetToken;
var
  nPos: Integer;
begin
  nPos := SendMessage(Handle, cb_GetEditSel, 0, 0) div 65536;
  SetSeperators;
  if nPos <= nSep1 then Token := 1 else
  if nPos <= nSep2 then Token := 2
  else Token := 3;
end;

(*********************************************
Select the token the cursor is on.
*********************************************)
procedure TDateEdit97.SelectToken;
begin
  case Token of
    1: SendMessage(Handle, em_SetSel, 0, (nSep1 - 1) * 65536);
    2: SendMessage(Handle, em_SetSel, 0, (nSep1 + (nSep2 - 1) * 65536));
    3: SendMessage(Handle, em_SetSel, 0, nSep2 + ((Length(Text)) * 65536));
  end;
end;

(*********************************************
Clear the Date field
*********************************************)
procedure TDateEdit97.ClearDate (Sender: TObject);
begin
  Click;
  FDate := 0;
  Text := '';
end;

(*********************************************
Clear Button
*********************************************)
procedure TDateEdit97.SetCanClear (value : boolean);
begin
  if FCanClear<>value
  then begin
    FCanClear := value;
    FBtnClear.Visible := value;
    Update;
  end;
end;

{-------------------------------------------------------------}
{-------------------- TDbDateEdit97 --------------------------}
{-------------------------------------------------------------}
constructor TDbDateEdit97.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  inherited ReadOnly := True;
  FDataLink := TFieldDataLink.Create;
  FDataLink.Control := Self;
  FDataLink.OnDataChange := DataChange;
  FDataLink.OnEditingChange := EditingChange;
  FDataLink.OnUpdateData := UpdateData;
end;

destructor TDbDateEdit97.Destroy;
begin
  FDataLink.Free;
  FDataLink := nil;
  FCanvas.Free;
  inherited Destroy;
end;

procedure TDbDateEdit97.Notification(AComponent: TComponent; Operation: TOperation);
begin
  inherited Notification(AComponent, Operation);
  if (Operation = opRemove) and (FDataLink <> nil) and
     (AComponent = DataSource) then DataSource := nil;
end;

procedure TDbDateEdit97.SetDate(dtArg: TDateTime);
var
  FormattedDate: string;
begin
  inherited SetDate(dtArg);
  if Field.AsDateTime <> dtArg then
  begin
    Field.AsDateTime := dtArg;
    Modified := True;
  end;
end;

procedure TDbDateEdit97.KeyPress(var Key: char);
begin
  if ((Key in ['0'..'9']) or (Key=DateSeparator)) and
     (FDataLink.Field <> nil) and not FDataLink.Field.IsValidChar(Key) then
  begin
    MessageBeep(0);
    Key := #0;
  end;

  case Key of
    ^H, ^V, ^X, '0'..'9': begin
      FDataLink.Edit;
    end;

    #27: begin
      FDataLink.Reset;
      SelectAll;
      Key := #0;
    end;
  end;

  inherited KeyPress(Key);
end;

procedure TDbDateEdit97.KeyDown(var Key: Word; Shift: TShiftState);
begin
  inherited KeyDown(Key, Shift);
  if (Key = VK_DELETE) or ((Key = VK_INSERT) and (ssShift in Shift)) then
    FDataLink.Edit;
end;

procedure TDbDateEdit97.Click;
begin
  FDataLink.Edit;
  inherited Click;
  FDataLink.Modified;
end;

procedure TDbDateEdit97.SetFocused(Value: Boolean);
begin
  if FFocused <> Value then
  begin
    FFocused := Value;
    if (FAlignment <> taLeftJustify) then Invalidate;
    FDataLink.Reset;
  end;
end;

procedure TDbDateEdit97.Change;
begin
  FDataLink.Modified;
  inherited Change;
end;

function TDbDateEdit97.GetDataSource: TDataSource;
begin
  Result := FDataLink.DataSource;
end;

procedure TDbDateEdit97.SetDataSource(Value: TDataSource);
begin
  FDataLink.DataSource := Value;
end;

function TDbDateEdit97.GetDataField: string;
begin
  Result := FDataLink.FieldName;
end;

procedure TDbDateEdit97.SetDataField(const Value: string);
begin
  FDataLink.FieldName := Value;
end;

function TDbDateEdit97.GetReadOnly: Boolean;
begin
  Result := FDataLink.ReadOnly;
end;

procedure TDbDateEdit97.SetReadOnly(Value: Boolean);
begin
  FDataLink.ReadOnly := Value;
end;

function TDbDateEdit97.GetField: TField;
begin
  Result := FDataLink.Field;
end;

procedure TDbDateEdit97.DataChange(Sender: TObject);
begin
  if FDataLink.Field <> nil then
  begin
    if FAlignment <> FDataLink.Field.Alignment then
    begin
      Text := '';  {forces update}
      FAlignment := FDataLink.Field.Alignment;
    end;

    if FDataLink.Field.DataType = ftString then
      MaxLength := FDataLink.Field.Size else
      MaxLength := 0;

    if FFocused and FDataLink.CanModify then
      Text := FDataLink.Field.Text
    else
      Text := FDataLink.Field.DisplayText;
  end else
  begin
    FAlignment := taLeftJustify;
    MaxLength := 0;
    if csDesigning in ComponentState then
      Text := Name else Text := '';
  end;
end;

procedure TDbDateEdit97.EditingChange(Sender: TObject);
begin
  inherited ReadOnly := not FDataLink.Editing;
end;

procedure TDbDateEdit97.UpdateData(Sender: TObject);
begin
  try
    FDataLink.Field.Text := Text;
  except
  end;
end;

procedure TDbDateEdit97.WMPaste(var message: TMessage);
begin
  FDataLink.Edit;
  inherited;
end;

procedure TDbDateEdit97.WMCut(var message: TMessage);
begin
  FDataLink.Edit;
  inherited;
end;

procedure TDbDateEdit97.CMEnter(var message: TCMEnter);
begin
  SetFocused(True);
  inherited;
end;

procedure TDbDateEdit97.CMExit(var message: TCMExit);
begin
  try
    FDataLink.UpdateRecord;
  except
    SelectAll;
    SetFocus;
    raise;
  end;

  SetFocused(False);
  SetCursor(0);
  DoExit;
end;

procedure Register;
begin
  RegisterComponents('Freeware', [TDateEdit97]);
  RegisterComponents('Freeware', [TDbDateEdit97]);
end;

end.
